import logging
import os
import re

import sau
import sau.helpers

EIX_SYNC_PATH='/usr/bin/eix-sync'
RC_SERVICE_PATH='/sbin/rc-service'
SYSTEMCTL='/usr/bin/systemctl'
EMERGE_PATH='/usr/bin/emerge'
EQUERY_PATH='/usr/bin/equery'
EMAINT_PATH='/usr/sbin/emaint'
GRUB_MKCONFIG='/usr/sbin/grub-mkconfig'

# parsing output from eix -Ttnc
package_re = re.compile('^\[([^\]])\] ([^ ]*) \((.*)\): .*$')
# parsing version information from substrings of the above
slot_re = re.compile('^(\(~\))?([^\(]+)(\([^\)]+\))$')

def identify_service_from_bin(exe):
    log = logging.getLogger(sau.LOGNAME)

    with open('/proc/1/comm', 'r') as f:
        if f.readline().strip() == 'systemd':
            init_script_re = re.compile(r'[^/]*(.*)\.service$')
        else:
            init_script_re = re.compile(r'/etc/init\.d/(.*)')

    cmd = [ EQUERY_PATH, '-Cq', 'b', exe ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if ret != 0:
        raise sau.errors.UnknownServiceError("searching for owner of {} failed:".format(exe))
    
    pkg = out.strip()
    cmd = [ EQUERY_PATH, '-Cq', 'f', pkg ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if ret != 0:
        log.warning("listing files for package {} failed:".format(pkg))
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))
        return None
    matches = set()
    for line in out.splitlines():
        match = re.match(init_script_re, line)
        if match:
            matches.add(match.group(1))
    if len(matches) < 1:
        raise sau.errors.UnknownServiceError('Could not find any init script in package {}'.format(pkg))
    elif len(matches) > 1:
        raise sau.errors.UnknownServiceError('Found multiple init script in package {}'.format(pkg))
    else:
        return matches.pop()
    return None
    

def restart_service(service):
    log = logging.getLogger(sau.LOGNAME)
    with open('/proc/1/comm', 'r') as f:
        if f.readline().strip() == 'systemd':
            cmd = [ SYSTEMCTL, 'restart', service ]
        else:
            cmd = [ RC_SERVICE_PATH, service, 'restart' ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if ret != 0:
        log.warning("Restart of {} failed:".format(service))
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))
    else:
        log.info("restarted service {}".format(service))

def system_upgrade():
    log = logging.getLogger(sau.LOGNAME)
    log.debug('Gentoo "system_upgrade" is done at package upgrade stage; ignoring here...')
    return False

def _sync_portage():
    log = logging.getLogger(sau.LOGNAME)

    if os.path.exists(EIX_SYNC_PATH):
        cmd = [ EIX_SYNC_PATH, '-q' ]
        ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)
    else:
        cmd = [ EMERGE_PATH, '-q', '--sync' ]
        ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)

    if ret != 0:
        log.error("Portage sync failed:")
        for line in out.splitlines():
            log.error("stdout: {}".format(line))
        for line in err.splitlines():
            log.error("stderr: {}".format(line))
        raise sau.errors.UpgradeError(f'Sync command {cmd} failed')

    cmd = [ EMAINT_PATH, '-f', 'all' ]
    ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)

    if ret != 0:
        log.warning("emaint failed:")
        for line in out.splitlines():
            log.warning("stdout: {}".format(line))
        for line in err.splitlines():
            log.warning("stderr: {}".format(line))




def pkg_upgrade():
    log = logging.getLogger(sau.LOGNAME)
    conf = sau.config
    do_system_upgrade = conf.getboolean('default', 'do_system_upgrade', fallback=False)

    if conf.getboolean('default', 'do_reposync', fallback=True):
        _sync_portage()

    # [ebuild     U ] media-plugins/alsa-plugins-1.1.8 [1.1.6]
    pretend_re = re.compile(r'^\[(?:ebuild|binary) ([^\]]*)\] ([^ ]+?)-(\d[-\.\w]*)( \[[^\]]+\])?')
    ignore_re = re.compile(r'^(|.*caus.* rebuilds.*|.*scheduled for merge.*|.*waiting for lock on.*)$')
    
    default_version_sens = conf.getint('default', 'version_sensitivity', fallback=1)

    cmd = [ EMERGE_PATH, '--color', 'n', '-uDNpq', '--with-bdeps=y', '@world' ]
    ret, out, err = sau.helpers.exec_cmd(cmd)

    if not ret == 0:
        log.error('emerge pretend returned {}'.format(ret))
        for line in out.splitlines():
            log.error('stdout: {}'.format(line))
        for line in err.splitlines():
            log.error('stderr: {}'.format(line))
        raise sau.errors.UpgradeError(f'Failed to calculate upgrade path')

    do_rebuild = True
    do_grub = False
    for line in out.splitlines():
        if re.match(ignore_re, line):
            continue
        match = re.match(pretend_re, line)
        if not match:
            log.warning('No match on line: {}'.format(line))
            continue
        status = match.group(1)
        name = match.group(2)
        new = match.group(3)
        old = match.group(4)
        if not old:
            continue
        old = old.strip(' []')

        sens = conf.getint('packages', name, fallback=default_version_sens)
        common = sau.helpers.version_diff(new, old)
        if sens <= common:
            log.info('{} -- {} -> {} configured level {} <= pkg level {}'.format(name, old, new, sens, common))
        else:
            log.error('{} -- {} -> {} configured level {} > pkg level {}'.format(name, old, new, sens, common))
            do_rebuild = False
        if name.startswith('sys-kernel/'):
            if do_system_upgrade:
                do_grub = True
            else:
                raise sau.errors.UpgradeError(f"Kernel package {name} has an update, but system upgrade is disabled")

    if not do_rebuild:
        raise sau.errors.UpgradeError('Some packages require manual attention, did not upgrade')

    cmd = [ EMERGE_PATH, '--color', 'n', '-uDNq', '--with-bdeps=y', '@world' ]
    ret, out, err = sau.helpers.exec_cmd(cmd, timeout=72000)

    if ret != 0 or err:
        log.error('emerge returned {}'.format(ret))
        for line in out.splitlines():
            log.error('stdout: {}'.format(line))
        for line in err.splitlines():
            log.error('stderr: {}'.format(line))
        raise sau.errors.UpgradeError(f'Error during upgrade')
    else:
        log.info('upgrade complete')
        for line in out.splitlines():
            if line.startswith(' * '):
                log.warning(line)

    cmd = [ EMERGE_PATH, '--color', 'n', '-q', '@preserved-rebuild' ]
    ret, out, err = sau.helpers.exec_cmd(cmd, timeout=72000)

    if ret != 0 or err:
        log.error('preserved-rebuild returned {}'.format(ret))
        for line in out.splitlines():
            log.error('stdout: {}'.format(line))
        for line in err.splitlines():
            log.error('stderr: {}'.format(line))
        raise sau.errors.UpgradeError(f'Error during preserved-rebuild')
    else:
        log.info('preserved-rebuild complete')
        for line in out.splitlines():
            if line.startswith(' * '):
                log.warning(line)

    if conf.getboolean('default', 'do_depclean', fallback=False):
        cmd = [ EMERGE_PATH, '--color', 'n', '-q', '--depclean' ]
        ret, out, err = sau.helpers.exec_cmd(cmd, timeout=3600)
        if ret != 0 or err:
            log.error('depclean returned {}'.format(ret))
            for line in out.splitlines():
                log.error('stdout: {}'.format(line))
            for line in err.splitlines():
                log.error('stderr: {}'.format(line))
            raise sau.errors.UpgradeError(f'Error during depclean')
        else:
            log.info('depclean complete')
            for line in out.splitlines():
                if line.startswith(' * '):
                    log.warning(line)

    if do_grub and os.path.exists(GRUB_MKCONFIG):
        keep_kernels = conf.getint('default', 'keep_kernels', fallback=4)
        if keep_kernels < 1:
            log.error('keep_kernels cannot be less than one; falling back to default')
            keep_kernels = 4
        for root, dirs, files in os.walk('/boot'):
            for sysfile in ['config', 'initramfs', 'System.map', 'vmlinuz']:
                match = sorted(
                        [f for f in files if f.startswith(f'{sysfile}-')],
                        reverse=True)
                for f in match[keep_kernels:]:
                    log.debug(f"Removing old kernel file {f}")
                    os.remove(os.path.join(root, f))
            break

        cmd = [ GRUB_MKCONFIG, '-o', '/boot/grub/grub.cfg' ]
        ret, out, err = sau.helpers.exec_cmd(cmd)
        if ret != 0:
            log.error(f"grub-mkconfig returned {ret}:")
            for line in out.splitlines():
                log.error('stdout: {}'.format(line))
            for line in err.splitlines():
                log.error('stderr: {}'.format(line))
            raise sau.errors.UpgradeError(f'Failed to reconfiugre grub')
        else:
            log.info("grub reconfigured")
            return True
